#include "rainbow/rdf-resource.h"
#include <libxml++/libxml++.h>
#include <postgresql/libpq-fe.h>
#include <iostream>

using namespace std;
using namespace Rainbow;

static PGconn *db = 0;

static void continue_crawl ();
static void connect_to_database ();

int main ()
{
	Glib::thread_init();

	connect_to_database();

	// search rainbow-builder for new uri's
	//PGresult *result = PQexec(db, "insert into webcrawler (select  from  where not in (select url from webcrawler))");

	// search recommendation returns for new uri's

	continue_crawl();

	Glib::RefPtr<Glib::MainLoop> ml = Glib::MainLoop::create();
        ml->run();

	return 0;
}

void connect_to_database ()
{
 again:
	if (db && PQstatus(db) == CONNECTION_OK)
		return;
	db = PQconnectdb("host=localhost dbname=gnomoradio-search");

	if (!db) {
		sleep(600);
		goto again;
	}
}

std::string add_slashes (const std::string &input)
{
	char *s = new char[input.size() * 2 + 1];
	PQescapeString(s, input.c_str(), input.size());
	string retval(s);
	delete[] s;
	return retval;
}

class Crawl : public SigC::Object
{
public:
	Crawl (const std::string &uri);
	~Crawl ();

	void do_it (xmlpp::Element *root, ref_ptr<RdfResource> resource);
private:
	bool do_artist (xmlpp::Element *artist, ref_ptr<RdfResource> resource);
	bool do_album (xmlpp::Element *artist, ref_ptr<RdfResource> resource);
	bool do_recording (xmlpp::Element *artist, ref_ptr<RdfResource> resource);

	std::string base_uri;
};

static void continue_crawl ()
{
	string uri;

	for (;;) {
		PGresult *res = PQexec(db, "select url from base_urls where last_attempt < current_timestamp - interval '3 hours' and last_update < current_timestamp - interval '7 days' limit 1");
		if (PQresultStatus(res) == PGRES_TUPLES_OK
		    && PQntuples(res))
			uri = PQgetvalue(res, 0, 0);
		PQclear(res);

		if (uri.size())
			break;
		else
			exit(0); //sleep(1800);

		connect_to_database();
	}

	PQclear(PQexec(db, string("update base_urls set last_attempt = current_timestamp where url='" + add_slashes(uri) + "'").c_str()));

	new Crawl (uri);
}

Crawl::Crawl (const std::string &uri)
	: base_uri(uri)
{
	RdfResource::get_and_do(base_uri, slot(*this, &Crawl::do_it));
}

Crawl::~Crawl ()
{
	continue_crawl();
}

void Crawl::do_it (xmlpp::Element *root, ref_ptr<RdfResource> resource)
{
	if (!root) {
		delete this;
		return;
	}

	PQclear(PQexec(db, string("update base_urls set last_update = current_timestamp where url='" + add_slashes(base_uri) + "'").c_str()));

	// We will scan for artists, albums, and then songs, because they reference each other in that order

	xmlpp::Node::NodeList children = root->get_children();
	xmlpp::Node::NodeList::iterator i;

	for (i = children.begin(); i != children.end(); ++i) {
		xmlpp::Element *el = dynamic_cast<xmlpp::Element*>(*i);
		if (el && el->get_name() == "Agent")
			do_artist(el, resource);
	}

#if 0
	for (i = children.begin(); i != children.end(); ++i) {
		xmlpp::Element *el = dynamic_cast<xmlpp::Element*>(*i);
		if (el && el->get_name() == "Album")
			do_album(el, resource);
	}
#endif

	for (i = children.begin(); i != children.end(); ++i) {
		xmlpp::Element *el = dynamic_cast<xmlpp::Element*>(*i);
		if (el && el->get_name() == "Recording")
			do_recording(el, resource);
	}

	delete this;
}

bool Crawl::do_recording (xmlpp::Element *recording, ref_ptr<RdfResource> resource)
{
	if (!recording)
		return false;

	xmlpp::Attribute *recording_id = recording->get_attribute("ID");
	if (!recording_id)
		return false;
	string url = base_uri + "#" + recording_id->get_value();

	string title;
	Glib::ustring creator;

	xmlpp::Node::NodeList children = recording->get_children();
	for (xmlpp::Node::NodeList::iterator i = children.begin(); i != children.end(); ++i) {
		xmlpp::Element *el = dynamic_cast<xmlpp::Element*>(*i);
		if (!el)
			continue;
		xmlpp::TextNode *content = el->get_child_text();
		xmlpp::Attribute *resource_link = el->get_attribute("resource");
		if (el->get_name() == "title") {
			if (content)
				title = content->get_content();
		} else if (el->get_name() == "creator") {
			if (resource_link)
				creator = RdfResource::absolute_uri(resource_link->get_value(), base_uri);
		}
	}

	// FIXME: make sure song is free

	if (!creator.size())
		return false; // no author
	PGresult *res;
	res = PQexec(db, string("select id from artists where url='" + add_slashes(creator) + "'").c_str());
	if (PQresultStatus(res) != PGRES_TUPLES_OK
		|| PQntuples(res) == 0) {
		PQclear(res);
		Glib::ustring tmp;
		RdfResource::remove_tag_from_uri(creator, tmp);
		PQclear(PQexec(db, string("insert into base_urls (url) values ('" + add_slashes(creator) + "')").c_str()));
		return false;
	}
	string db_artist_index;
	db_artist_index = PQgetvalue(res, 0, 0);
	PQclear(res);

	res = PQexec(db, string("select id from songs where url='" + add_slashes(url) + "'").c_str());
	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
		PQclear(res);
		return false;
	}
	bool in_db = PQntuples(res);
	string db_index;
	if (in_db)
		db_index = PQgetvalue(res, 0, 0);
	PQclear(res);

	if (title.size()) {
		if (in_db) {
			PQclear(PQexec(db, string("update songs set name='" + add_slashes(title) + "', artist='" + db_artist_index + "' where id='" + db_index + "'").c_str()));
		} else {
			PQclear(PQexec(db, string("insert into songs (url, name, artist) values ('" + add_slashes(url) + "', '" + add_slashes(title) + "', '" + db_artist_index + "')").c_str()));
		}
	} else {
		if (in_db) {
			PQclear(PQexec(db, string("delete from songs where id='" + db_index + "'").c_str()));
		}
	}

	return true;
}

bool Crawl::do_artist (xmlpp::Element *artist, ref_ptr<RdfResource> resource)
{
	if (!artist)
		return false;

	xmlpp::Attribute *artist_id = artist->get_attribute("ID");
	if (!artist_id)
		return false; // it has no name
	string url = base_uri + "#" + artist_id->get_value();

	string title, website;

	xmlpp::Node::NodeList children = artist->get_children();
	for (xmlpp::Node::NodeList::iterator i = children.begin(); i != children.end(); ++i) {
		xmlpp::Element *el = dynamic_cast<xmlpp::Element*>(*i);
		if (!el)
			continue;
		xmlpp::TextNode *content = el->get_child_text();
		xmlpp::Attribute *resource_link = el->get_attribute("resource");
		if (el->get_name() == "title") {
			if (content)
				title = content->get_content();
		} else if (el->get_name() == "website") {
			if (resource_link)
				website = RdfResource::absolute_uri(resource_link->get_value(), base_uri);
		}
	}

	PGresult *res = PQexec(db, string("select id from artists where url='" + add_slashes(url) + "'").c_str());
	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
		PQclear(res);
		return false;
	}
	bool in_db = PQntuples(res);
	string db_index;
	if (in_db)
		db_index = PQgetvalue(res, 0, 0);
	PQclear(res);

	if (title.size()) {
		if (in_db) {
			PQclear(PQexec(db, string("update artists set name='" + add_slashes(title) + "', website='" + add_slashes(website) + "' where id='" + db_index + "'").c_str()));
		} else {
			PQclear(PQexec(db, string("insert into artists (url, name, website) values ('" + add_slashes(url) + "', '" + add_slashes(title) + "', '" + add_slashes(website) + "')").c_str()));
		}
	} else {
		if (in_db) {
			PQclear(PQexec(db, string("delete from artists where id='" + db_index + "'").c_str()));
		}
	}

	return true;
}
